"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const js_yaml_1 = __importDefault(require("js-yaml"));
const ts_json_schema_generator_1 = require("ts-json-schema-generator");
const lodash_1 = require("lodash");
class ServerlessOpenapiTypeScript {
    constructor(serverless, options) {
        var _a;
        this.serverless = serverless;
        this.options = options;
        this.assertPluginOrder();
        this.initOptions(options);
        this.functionsMissingDocumentation = [];
        if (!((_a = this.serverless.service.custom) === null || _a === void 0 ? void 0 : _a.documentation)) {
            this.log(`Disabling OpenAPI generation for ${this.serverless.service.service} - no 'custom.documentation' attribute found`);
            this.disable = true;
            delete this.serverless.pluginManager.hooks['openapi:generate:serverless'];
        }
        if (!this.disable) {
            this.hooks = {
                'before:openapi:generate:serverless': this.populateServerlessWithModels.bind(this),
                'after:openapi:generate:serverless': this.postProcessOpenApi.bind(this)
            };
        }
    }
    initOptions(options) {
        this.options = options || {};
        this.typescriptApiModelPath = this.options.typescriptApiPath || 'api.d.ts';
        this.tsconfigPath = this.options.tsconfigPath || 'tsconfig.json';
    }
    assertPluginOrder() {
        if (!this.serverless.pluginManager.hooks['openapi:generate:serverless']) {
            throw new Error('Please configure your serverless.plugins list so serverless-openapi-typescript will be listed AFTER @conqa/serverless-openapi-documentation');
        }
    }
    get functions() {
        return this.serverless.service.functions || {};
    }
    log(msg) {
        this.serverless.cli.log(`[serverless-openapi-typescript] ${msg}`);
    }
    populateServerlessWithModels() {
        return __awaiter(this, void 0, void 0, function* () {
            this.log('Scanning functions for documentation attribute');
            Object.keys(this.functions).forEach(functionName => {
                var _a, _b;
                (_b = (_a = this.functions[functionName]) === null || _a === void 0 ? void 0 : _a.events) === null || _b === void 0 ? void 0 : _b.forEach((event) => {
                    const httpEvent = event.http;
                    if (httpEvent) {
                        if (httpEvent.documentation) {
                            this.log(`Generating docs for ${functionName}`);
                            this.setModels(httpEvent, functionName);
                            const paths = (0, lodash_1.get)(httpEvent, 'request.parameters.paths', []);
                            const querystrings = (0, lodash_1.get)(httpEvent, 'request.parameters.querystrings', {});
                            [
                                { params: paths, documentationKey: 'pathParams' },
                                { params: querystrings, documentationKey: 'queryParams' }
                            ].forEach(({ params, documentationKey }) => {
                                this.setDefaultParamsDocumentation(params, httpEvent, documentationKey);
                            });
                        }
                        else if (httpEvent.documentation !== null && !httpEvent.private) {
                            this.functionsMissingDocumentation.push(functionName);
                        }
                    }
                });
            });
            this.assertAllFunctionsDocumented();
        });
    }
    assertAllFunctionsDocumented() {
        if (!(0, lodash_1.isEmpty)(this.functionsMissingDocumentation)) {
            throw new Error(`Some functions have http events which are not documented:
         ${this.functionsMissingDocumentation}
         
        Please add a documentation attribute. 
        If you wish to keep the function undocumented, please explicitly set 
        documentation: ~
         
        `);
        }
    }
    setDefaultParamsDocumentation(params, httpEvent, documentationKey) {
        Object.entries(params).forEach(([name, required]) => {
            httpEvent.documentation[documentationKey] = httpEvent.documentation[documentationKey] || [];
            const documentedParams = httpEvent.documentation[documentationKey];
            const existingDocumentedParam = documentedParams.find(documentedParam => documentedParam.name === name);
            if (existingDocumentedParam && typeof existingDocumentedParam.schema === 'string') {
                existingDocumentedParam.schema = this.generateSchema(existingDocumentedParam.schema);
            }
            const paramDocumentationFromSls = {
                name,
                required,
                schema: { type: 'string' }
            };
            if (!existingDocumentedParam) {
                documentedParams.push(paramDocumentationFromSls);
            }
            else {
                Object.assign(paramDocumentationFromSls, existingDocumentedParam);
                Object.assign(existingDocumentedParam, paramDocumentationFromSls);
            }
        });
    }
    setModels(httpEvent, functionName) {
        const definitionPrefix = `${this.serverless.service.custom.documentation.apiNamespace}.${(0, lodash_1.upperFirst)(functionName)}`;
        const method = httpEvent.method.toLowerCase();
        switch (method) {
            case 'delete':
                (0, lodash_1.set)(httpEvent, 'documentation.methodResponses', [{ statusCode: 204, responseModels: {} }]);
                break;
            case 'patch':
            case 'put':
            case 'post':
                const requestModelName = `${definitionPrefix}.Request.Body`;
                this.setModel(`${definitionPrefix}.Request.Body`);
                (0, lodash_1.set)(httpEvent, 'documentation.requestModels', { 'application/json': requestModelName });
                (0, lodash_1.set)(httpEvent, 'documentation.requestBody', { description: '' });
            case 'get':
                const responseModelName = `${definitionPrefix}.Response`;
                this.setModel(`${definitionPrefix}.Response`);
                (0, lodash_1.set)(httpEvent, 'documentation.methodResponses', [
                    {
                        statusCode: 200,
                        responseBody: { description: '' },
                        responseModels: { 'application/json': responseModelName }
                    }
                ]);
        }
        const queryParamModel = `${definitionPrefix}.Request.QueryParams`;
        try {
            this.setModel(queryParamModel);
        }
        catch (e) {
            this.log(`Skipped generation of "${queryParamModel}" - model is missing - will be using the default query param of type string`);
        }
        const pathParamModel = `${definitionPrefix}.Request.PathParams`;
        try {
            this.setModel(pathParamModel);
        }
        catch (e) {
            this.log(`Skipped generation of "${pathParamModel}" - model is missing - will be using the default path param of type string`);
        }
    }
    postProcessOpenApi() {
        const outputFile = this.serverless.processedInput.options.output;
        const openApi = js_yaml_1.default.load(fs_1.default.readFileSync(outputFile));
        this.patchOpenApiVersion(openApi);
        this.tagMethods(openApi);
        fs_1.default.writeFileSync(outputFile, outputFile.endsWith('json') ? JSON.stringify(openApi, null, 2) : js_yaml_1.default.dump(openApi));
    }
    patchOpenApiVersion(openApi) {
        this.log(`Setting openapi version to 3.1.0`);
        openApi.openapi = '3.1.0';
        return openApi;
    }
    tagMethods(openApi) {
        var _a;
        const tagName = openApi.info.title;
        openApi.tags = [
            {
                name: tagName,
                description: openApi.info.description
            }
        ];
        const customTags = (_a = this.serverless.service.custom.documentation) === null || _a === void 0 ? void 0 : _a.tags;
        if (customTags)
            openApi.tags = openApi.tags.concat(customTags);
        Object.values(openApi.paths).forEach(path => {
            Object.values(path).forEach(method => {
                var _a, _b, _c;
                const httpEvent = (_b = (_a = this.functions[method.operationId]) === null || _a === void 0 ? void 0 : _a.events) === null || _b === void 0 ? void 0 : _b.find((e) => e.http);
                const http = httpEvent.http;
                if ((_c = http.documentation) === null || _c === void 0 ? void 0 : _c.tag) {
                    method.tags = [http.documentation.tag];
                }
                else {
                    method.tags = [tagName];
                }
            });
        });
    }
    setModel(modelName) {
        (0, lodash_1.mergeWith)(this.serverless.service.custom, {
            documentation: {
                models: [{ name: modelName, contentType: 'application/json', schema: this.generateSchema(modelName) }]
            }
        }, (objValue, srcValue) => {
            if ((0, lodash_1.isArray)(objValue)) {
                return objValue.concat(srcValue);
            }
        });
    }
    generateSchema(modelName) {
        this.log(`Generating schema for ${modelName}`);
        this.schemaGenerator =
            this.schemaGenerator ||
                (0, ts_json_schema_generator_1.createGenerator)({
                    path: this.typescriptApiModelPath,
                    tsconfig: this.tsconfigPath,
                    type: `*`,
                    expose: 'export',
                    skipTypeCheck: true,
                    topRef: false
                });
        return this.schemaGenerator.createSchema(modelName);
    }
}
exports.default = ServerlessOpenapiTypeScript;
//# sourceMappingURL=serverless-openapi-typescript.js.map